package RFC::Simple;
#$Id: Simple.pm,v 1.81 2015/11/11 09:19:55 sqko Exp $

#################################################################
#
#
#       RFC::Simple
#
# This class defines simple dataloging (readonly) channel setup
# To obtain an instance of this class call one of the constructors 
# like:
#
# $id = RFC::Simple->new($name,$rig)
#
# In case of an RFC::keithley or RFC::SimpleICP instance, it can 
# also be obtained by 
#
# $id = RFC::Keithley->new($name,$rig,$channel)
#
# or similarly for a RFC::SimpleICP instance.
#  
# The reason for the last option is to allow the instantiation 
# of a device without a separate setup section.
#
# The RFC::Simple module defines no new member functions.
# However all simplechannel devices have the configuration option
# of using a scaling factor (effectively a number which is multiplied
# with the raw physical value). For instance this enables a value
# to be reported in mV even though it is measured in Volt by setting
# the scaling factor to 1000. 
#
##################################################################

use strict;
use warnings;
use RFC::BaseDevice;
use RFC::Cache;
our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    bless $self,$class;
    $self->{'_type'} = 'SimpleChannel';
    $self->{'_mode'} = 'Readonly';
    $self->{'_setup_tags'} = ['mode','channel','factor','title','show_plot','use_cache'];
    $self->{'_desc'} = {
	'mode' => 'Device control mode',
	'channel' => 'Channel number of input device',
	'factor' => 'Scaling factor',
	'title' => 'Optional title',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
	'type' => 'Keithley',
        'mode' => 'Readonly',
        'channel' => 0,
        'factor' => 1,
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_values_mode'} = ['Readonly'];
    $self->{'_channel'} = undef;
    my $test = shift;
    $self->{'_channel'} = $test if ($test);
    $self->{'_section'} = 'channel_'.$self->name;
    $self->{'_title'} = $self->get_conf('title');
    $self->{'_readonly'} = 1;
    return $self;
}


sub read_nocache
{
    my $self = shift;
    $self->die($self->type." is a base class and does not impelemnt id->read_nocache(), only the derived classes does so!");
    return undef;
}

sub readstring
{
    my $self = shift;
    if ($self->is_caching())
    {
	return $self->SUPER::readstring(@_);
    }
    return undef if ($self->readstring_ignore);
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    $self->read($data);
    my $res = $data->get($self->type,$self->{'_channel'});
    if ($res)
    {
	my @rep = split(/\s+/,$res);
	$res = join(' ',$self->name,$rep[0],$rep[1],$rep[$#rep]*$self->{'_factor'});
	$self->writedebug($res) if ($self->debug > 2);
	return $res;
    }
    else
    {
	my @t = $self->get_time();
	my $t = join(':',@t);
        $self->writedebug("No data could be read from device '".$self->name."' using dummy value!") if ($self->debug);
	my $res = join(' ',$self->name,$t,time,-32768);
	return $res;
    }
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Input type: '.$self->type;
    push @arr,'Channel: '.$self->{'_channel'};
    push @arr,'Factor: '.$self->{'_factor'};
    return join(', ',@arr);
}

package RFC::Keithley;
our @ISA = qw(RFC::Simple);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'Keithley';
    $self->{'_title'} = $self->get_conf('title');
    $self->{'_channel'} = $self->get_conf('channel') unless (defined($self->{'_channel'}));
    $self->{'_channel'} = $self->get_default('channel') unless ($self->{'_channel'});
    $self->{'_factor'} = $self->get_conf('factor') unless (defined($self->{'_factor'}));
    $self->{'_factor'} = $self->get_default('factor') unless ($self->{'_factor'});
    $self->{'_use_cache'} = $self->get_conf('use_cache') unless (defined($self->{'_use_cache'}));
    $self->{'_use_cache'} = $self->get_default('use_cache') unless ($self->{'_use_cache'});
    $self->{'_desc'} = {
	'mode' => 'Device control mode',
	'channel' => 'Channel number of input device, format: X:YZZ, X gpib address, Y board number, ZZ channel number on board',
	'factor' => 'Scaling factor',
	'title' => 'Optional title',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    bless $self,$class;
    return $self;    
}

sub test()
{
    my $self = shift;
    my $ok = 0;
    my $mrep = SocketClient::GPIB_client_raw('K',$self->{'_channel'});
    chomp($mrep);
    if ($mrep =~ /([\w\s\.\:\+\-\;]+)/)
    {
	$mrep = $1;
	my @rep = split (/\s+/, $mrep );    # split svar op i enkeltdele, tid, vaerdi, etc
	if ($self->{'_channel'} eq $rep[2] && $mrep !~ /error/i)
	{
	    $ok = 1;
	}
    }
    return $ok;
}

sub read_nocache
{
    my $self = shift;
    my $res = $self->read_raw(@_);
    if (defined($res))
    {
	my @rep = split(/\s+/,$res);
	$self->writedebug("Channel ".$self->name." was read to $rep[$#rep]") if ($self->debug > 1);
	$rep[$#rep] = $rep[$#rep]*$self->{'_factor'};
	my $str = join(' ',time,$self->name,join(' ',@rep));
	$self->rig->kanaldata($str) unless ($self->readstring_ignore);
	$self->set_cache($rep[$#rep]);
	$self->{'_last_read'} = Time::HiRes::time;
	return $rep[$#rep];
    }
    else
    {
	$self->writedebug("Channel ".$self->name." returned undefined on read!") if ($self->debug);
    }
    return undef;
}

sub test_string
{
    my $self = shift;
    my $res = $self->read_raw(@_);
    $res =~ s/[\s\t]*$//;
    if ($res)
    {
	my @l = split(/\s+/,$res);
	$l[-1] = $l[-1]*$self->{'_factor'};
	$res = join(' ',@l);
    }
    return "Simple channel ".$self->name,": $res\n";
}

sub read_raw
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my $cached = $data->get($self->type,$self->{'_channel'});
    if (defined($cached))
    {
	$self->writedebug("Using already measured value for ".$self->name) if ($self->debug > 3); 
	$res = $cached;
    }
    else
    {
	my $ok = 0;
	my $try = 5;
	$self->writedebug("Measuring value for ".$self->name) if ($self->debug > 3); 
	while (!$ok && $try--)
	{
	    my $mrep = SocketClient::GPIB_client_raw('K',$self->{'_channel'}); # svar fra maaling fra keithley
	    chomp($mrep) if defined($mrep);                   # fjern linieskift i slut af svar
	    $self->writedebug("Response: $mrep") if ($self->debug > 3);
	    next unless ($mrep);
	    $mrep =~ /([\w\s\.\:\+\-\;]+)/;
	    $mrep = $1;
	    my @rep = split (/\s+/, $mrep );    # split svar op i enkeltdele, tid, vaerdi, etc
	    if ($self->{'_channel'} eq $rep[2])
	    {
		$res = $mrep;
		$ok = 1;
	    }
	}
	$data->set($self->type,$self->{'_channel'},$res);
    }
    unless (defined($res))
    {
	$self->writedebug("No data found for channel ".$self->name);
	return undef;
    }
    chomp $res;
    return $res;
}

package RFC::KeithleyMOM;
our @ISA = qw(RFC::Simple);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    bless $self,$class;
    $self->{'_type'} = 'KeithleyMicroohmmeter';
    $self->{'_title'} = $self->get_conf('title');
    $self->{'_mode'} = 'Readonly';
    $self->{'_setup_tags'} = ['mode','address','range','dry_circuit','relative','polarity','drive','factor','title','show_plot','use_cache'];
    $self->{'_desc'} = {
	'mode' => 'Device control mode',
	'address' => 'GPIB address',
	'range' => 'Measurement range',
	'dry_circuit' => 'Use dry circuit measurement mode',
	'relative' => 'Use relative measruement mode',
	'plarity' => 'Select polarity for measurement',
	'drive' => 'Measurement drive mode',
	'title' => 'Optional title',
	'factor' => 'Scaling factor',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
	'type' => 'Keithley580',
        'mode' => 'Readonly',
        'address' => 1,
        'factor' => 1,
	'range' => 0,
	'dry_circuit' => 0,
	'relative' => 0,
	'polarity' => 0,
	'drive' => 1,
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_values_range'} = ['Auto','200m','2','20','200','2k','20k','200k'];
    $self->{'_values_dry_circuit'} = ['None','Enabled'];
    $self->{'_values_relative'} = ['Off','On'];
    $self->{'_values_mode'} = ['Readonly'];
    $self->{'_values_polarity'} = ['Pol+','Pol-'];
    $self->{'_values_drive'} = ['Pulsed','DC'];
    foreach my $key ('address','use_cache')
    {
        $self->{'_'.$key} = $self->get_conf($key);
        $self->{'_'.$key} = $self->get_default($key) unless ($self->{'_'.$key});
    }
    foreach my $key ('range','dry_circuit','relative','polarity','drive','factor')
    {
        $self->{'_'.$key} = $self->get_tag_index($key,$self->get_conf($key));
        $self->{'_'.$key} = $self->get_default($key) unless (defined($self->{'_'.$key}));
    }
    return $self;
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Input type: '.$self->type;
    push @arr,'Address: '.$self->{'_address'};
    push @arr,'Factor: '.$self->{'_factor'};
    my $cmd = 'D'.$self->{'_drive'}.'C'.$self->{'_dry_circuit'}.'R'.$self->{'_range'}.'P'.$self->{'_polarity'}.'Z'.$self->{'_relative'}.'T1X';
    push @arr,'Setupstring: '.$cmd;
    return join(', ',@arr);
}

sub test()
{
    my $self = shift;
    my $ok = 0;
    my $cmd = 'D'.$self->{'_drive'}.'C'.$self->{'_dry_circuit'}.'R'.$self->{'_range'}.'P'.$self->{'_polarity'}.'Z'.$self->{'_relative'}.'T1X';
    my $mrep = SocketClient::GPIB_client_raw('c',$self->{'_address'},$cmd);
    chomp($mrep) if (defined($mrep));
    if (defined($mrep) && $mrep =~ /(\d+)/)
    {
	$ok = 1;
    }
    return $ok;
}

sub get_tag_index()
{
    my $self = shift;
    my $tag = shift;
    my $value = shift;
    return undef unless defined($value);
    if (exists($self->{'_values_'.$tag}))
    {
	my @l = @{$self->{'_values_'.$tag}};
	for (my $x = 0; $x < @l; $x++)
	{
	    return $x if ($l[$x] eq $value)
	}
    }
    return undef;
}

sub readstring
{
    my $self = shift;
    return undef if ($self->readstring_ignore);
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my $val = $self->read($data);
    my @t = $self->get_time();
    my $t = join(':',@t);
    my $res = join(' ',$self->name,$t,time,$val);
    $self->writedebug($res) if ($self->debug > 1);
    return $res;
}

sub read_nocache
{
    my $self = shift;
    my $res = $self->read_raw(@_);
    my $value = -32768;
    $res = '' unless defined($res);
    if ($res =~ /([\+\-\.\deE]+)[\s\t]*$/)
    {
	$value = $1+0;
    }
    $self->writedebug("Channel ".$self->name." was read to $value") if ($self->debug > 1);
    my $str = time.' '.$self->name.' '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    $self->set_cache($value * $self->{'_factor'});
    $self->{'_last_read'} = Time::HiRes::time;
    return $value * $self->{'_factor'};
}

sub read_raw
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my $cmd = 'D'.$self->{'_drive'}.'C'.$self->{'_dry_circuit'}.'R'.$self->{'_range'}.'P'.$self->{'_polarity'}.'Z'.$self->{'_relative'}.'T1X';
    my $cached = $data->get($self->type,$self->{'_address'},$cmd);
    if (defined($cached))
    {
	$self->writedebug("Using already measured value for ".$self->name) if ($self->debug > 3); 
	$res = $cached;
    }
    else
    {
	my $ok = 0;
	my $try = 5;
	$self->writedebug("Measuring value for ".$self->name) if ($self->debug > 3); 
	while (!$ok && $try--)
	{
	    my $mrep = SocketClient::GPIB_client_raw('c',$self->{'_address'},$cmd); # svar fra maaling fra keithley
	    $mrep = '' unless (defined($mrep));
	    chomp($mrep);   # fjern linieskift i slut af svar
	    $self->writedebug("Response: '$mrep'") if ($self->debug > 3);
	    next unless ($mrep);
	    if ($mrep =~ /([\w\s\.\:\+\-]+)/)
	    {
		$mrep = $1;
		$res = $mrep;
		$ok = 1;
	    }
	}
	if (defined($data))
	{
	    $data->set($self->type,$self->{'_address'},$cmd,$res);
	}
    }
    unless (defined($res))
    {
	$self->writedebug("No data found for channel ".$self->name);
	return undef;
    }
    chomp $res;
    if ($res =~ /([\+\-]*\d+[\d\.\+\-eE]+)/)
    {
	$res = $1;
    }
    return $res * $self->{'_factor'};
}

sub test_string
{
    my $self = shift;
    my $res = $self->read_raw(@_);
    return "Simple channel ".$self->name,": $res\n";
}



package RFC::SimpleICP;
our @ISA = qw(RFC::Simple);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    bless $self,$class;
    $self->{'_type'} = 'SimpleICP';
    $self->{'_setup_tags'} = ['mode','tty','address','channel','factor','title','show_plot','use_cache','CRC'];
    $self->{'_desc'} = {
	'mode' => 'Device control mode',
	'tty' => 'Serial device for controling device (eg ttyS0)',
	'address' => 'Device address',
	'channel' => 'Channel number on device (Note 0-based)',
	'title' => 'Optional title',
	'factor' => 'Scaling factor',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
	'CRC' => 'Determines if checksum is to be used for serial communication',
    };
    $self->{'_default'} = {
	'type' => 'ICP',
        'mode' => 'Readonly',
	'tty' => 'ttyS0',
	'address' => 1,
	'factor' => 1,
        'channel' => 0,
        'show_plot'=>'Yes',
        'use_cache'=>'No',
	'CRC'=>'No',
    };
    foreach my $key ('tty','address','channel','factor','use_cache','CRC')
    {
        $self->{'_'.$key} = $self->get_conf($key);
        $self->{'_'.$key} = $self->get_default($key) unless ($self->{'_'.$key});
    }
    $self->{'_title'} = $self->get_conf('title');
    $self->{'_tty'} = shift if (@_);
    $self->{'_address'} = shift if (@_);
    $self->{'_channel'} = shift if (@_);
    $self->{'_values_CRC'} = ['No','Yes'];
    $self->{'_ui_fn_names'} = ['query_settings'];
    $self->{'_ui_fn_desc'} = {
        'query_settings' => 'Get and store the settings which can be queried from the module',
    };
    return $self;    
}

sub query_settings
{
    my $self = shift;
    my $u = '';
    my @list = ();
    $self->rig->begin();
    eval {
        my $str = $self->config;
	CORE::die "Could not get configuration from module $self\n" unless ($str);
        $self->change_cv('config',$str);
        push @list,"Config: $str";
        if ($str && $str =~ /\!\w{6}(\w)\w/)
        {
            my $d = $1;
            if ($d & 4)
            {
                $self->change_cv('CRC','Yes');
            }
            else
            {
                $self->change_cv('CRC','No');
            }
        }
        $self->rig->commit();
    };
    if ($@)
    {
        $self->rig->rollback();
        push @list,'Error: '.$@;
    }
    return join("\n",@list);
}

sub config()
{
    my $self = shift;
    my $res;
    my $cmd = sprintf("\$%02d2",$self->{'_address'});
    if ($self->{'_CRC'} eq 'Yes')
    {
        my $c = 'icp_raw_crc';
        $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        ### Try 2 times to get configuration ###
        unless (defined($res) && $res =~ /\w/)
        {
            select(undef,undef,undef,0.5);
            $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        }
        $c = 'icp_raw';
        ### Try 2 times more without checksum to get configuration ###
        unless (defined($res) && $res =~ /\w/)
        {
            select(undef,undef,undef,0.5);
            $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        }
        unless (defined($res) && $res =~ /\w/)
        {
            select(undef,undef,undef,0.5);
            $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        }
    }
    else
    {
        my $c = 'icp_raw';
        $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        ### Try 2 times to get configuration ###
        unless (defined($res) && $res =~ /\w/)
        {
            select(undef,undef,undef,0.5);
            $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        }
        ### Try 2 times more with checksum to get configuration ###
        $c = 'icp_raw_crc';
        unless (defined($res) && $res =~ /\w/)
        {
            select(undef,undef,undef,0.5);
            $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        }
        unless (defined($res) && $res =~ /\w/)
        {
            select(undef,undef,undef,0.5);
            $res = SocketClient::serial_client_raw($self->{'_tty'},$c,$cmd);
        }
    }
    chomp $res if (defined($res));
    return $res;
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Input type: '.$self->type;
    push @arr,'tty: '.$self->{'_tty'};
    push @arr,'Address: '.$self->{'_address'};
    push @arr,'Channel: '.$self->{'_channel'};
    return join(', ',@arr);
}

sub test()
{
    my $self = shift;
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    my $cmd = 'icpmultiread';
    $cmd = 'icpmultiread_crc' if ($self->{'_CRC'} eq 'Yes');
    my $mrep = SocketClient::serial_client_raw($tty,$cmd,$addr);
    return 0 unless ($mrep);
    my @r = split("\n",$mrep);
    $res = $r[$self->{'_channel'}] if ($self->{'_channel'} < @r);
    if (defined($res))
    {
	return 1;
    }
    return 0;
}

sub test_string()
{
    my $self = shift;
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    my $cmd = 'icpmultiread';
    $cmd = 'icpmultiread_crc' if ($self->{'_CRC'} eq 'Yes');
    my $mrep = SocketClient::serial_client_raw($tty,$cmd,$addr);
    return 0 unless ($mrep);
    my @r = split("\n",$mrep);
    $res = $r[$self->{'_channel'}] if ($self->{'_channel'} < @r);
    return $res;
}


sub read_nocache
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    my $cached = $data->get($tty,'icp',$addr,$self->{'_channel'});
    if ($cached)
    {
	$self->writedebug("Using already measured value for ".$self->name) if ($self->debug > 3); 
        $res = $cached;
    }
    else
    {
	$self->writedebug("Measuring value for ".$self->name." on tty $tty address $addr") if ($self->debug > 3);
	my $cmd = 'icpmultiread';
	$cmd = 'icpmultiread_crc' if ($self->{'_CRC'} eq 'Yes'); 
	my $mrep = SocketClient::serial_client_raw($tty,$cmd,$addr);
	unless ($mrep)
	{
	    select(undef,undef,undef,0.3);
	    $self->writedebug("No response, retrying...") if ($self->debug > 3);
	    $mrep = SocketClient::serial_client_raw($tty,$cmd,$addr);
	}
	$self->writedebug("Response: $mrep") if ($self->debug > 3);
	if ($mrep)
	{
	    my @r = split("\n",$mrep);
	    $res = $r[$self->{'_channel'}] if ($self->{'_channel'} < @r);
	    for (my $x = 0; $x < @r; $x++)
	    {
		$data->set($tty,'icp',$addr,$x,$r[$x]);
	    }
	}
    }
    $self->die("No data returned from device!") unless ($res);
    $self->writedebug("Channel ".$self->name." was read to $res") if ($self->debug > 1);
    my $str = time.' '.$self->name.' '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    $self->set_cache($res * $self->{'_factor'});
    $self->{'_last_read'} = Time::HiRes::time;
    return $res * $self->{'_factor'};
}

sub readstring
{
    my $self = shift;
    return undef if ($self->readstring_ignore);
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my $value = $self->read($data);
    $value = -32768 unless (defined($value) && $value =~ /\d/);
    my @t = $self->get_time();
    my $t = join(':',@t);
    my $res = join(' ',$self->name,$t,time,sprintf("%.6f",$value));
    $self->writedebug($res) if ($self->debug > 1);
    return $res;
}

package RFC::SimpleModbus;
our @ISA = qw(RFC::SimpleICP);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'SimpleModbus';
    $self->{'_title'} = $self->get_conf('title');
    $self->{'_desc'}{'channel'} = 'Modbus tag number for reading value';
    return $self;    
}

sub test()
{
    my $self = shift;
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    $res = SocketClient::serial_client_raw($tty,'modbus','f',$addr,$self->{'_channel'});
    return 1 if (defined($res));
    return 0;
}

sub test_string()
{
    my $self = shift;
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    return SocketClient::serial_client_raw($tty,'modbus','f',$addr,$self->{'_channel'});
}

sub read_nocache
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    my $type = $self->type;
    my $cached = $data->get($tty,$type,$addr,$self->{'_channel'});
    if (defined($cached))
    {
	$self->writedebug("Using already measured value for ".$self->name) if ($self->debug > 3); 
        $res = $cached;
    }
    else
    {
	$self->writedebug("Measuring value for ".$self->name." on tty $tty address $addr") if ($self->debug > 3); 
	$res = SocketClient::serial_client_raw($tty,'modbus','f',$addr,$self->{'_channel'});
	unless (defined($res))
	{
	    select(undef,undef,undef,0.3);
	    $self->writedebug("No response, retrying...") if ($self->debug > 3); 
	    $res = SocketClient::serial_client_raw($tty,'modbus','f',$addr,$self->{'_channel'});
	}
	$self->writedebug("Response: $res") if ($self->debug > 3);
	chomp $res if (defined($res));
	$data->set($tty,$type,$addr,$self->{'_channel'},$res) if (defined($res));
    }
    $self->die("No data returned from device!") unless ($res);
    $self->writedebug("Channel ".$self->name." was read to '$res'") if ($self->debug > 1);
    my $str = time.' '.$self->name.' '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    if (defined($res) && $res =~ /\d/)
    {
	$self->set_cache($res * $self->{'_factor'});
	$self->{'_last_read'} = Time::HiRes::time;
	return $res * $self->{'_factor'};
    }
    else
    {
	return -32768;
    }
}

return 1;
